# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a DPT database of chess games using the dptdb package.

This interface can be used if the dptdb package is installed.

The dptdb package is available as a source package from chesstab's home page.

Follow the instructions to build and install dptdb.

DPT is "Database Programmers Toolkit", which runs on Microsoft Windows only.

Support for DPT ceased in January 2015, but the source code of it's database
interface was available from DPT's home page until then, under a BSD licence,
and is provided in the dptdb package.
"""
