# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access an Sqlite3 database of chess games using the sqlite3 package.

This interface can be used if the sqlite3 package is installed.

The sqlite3 package is installed by default when Python is installed on some
platforms, but it has to be installed separately on others.
"""
