# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a Berkeley DB database of chess games using the bsddb3 package.

This interface can be used if the bsddb3 package is installed.

The bsddb3 package is available as a source package, from PyPI for example.

Follow the instructions to build and install bsddb3.  This includes installing
Berkeley DB.
"""
