# __init__.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a NoSQL database of chess games using the dbm.ndbm package.

This interface can be used if ndbm software is installed and available to the
dbm.ndbm package.
"""
